

##############################################################################
#
# srvconf.tcl --
#
#    Server startup and configuration procedure. This procedure is
#    executed at server startup only. Therefore only limited server
#    command-set is available.
#
#    See the file "license.txt" for information on usage and
#    redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
##############################################################################

#
# These two variables configure the control-port access information.
# You can declare only one control-port user.
# Normally, you do not need to access control port unless requested
# to do this by the Archiware tech-support.
#
# You may choose to set this file to be viewed by the system privileged
# user (root) in order to keep the control-port user password secret.
#
# The control port listens at the address (https-port+1000). If you setup
# the webport to 443 (the default setup) the control port will be set
# to 1443. Note: only localhost connects are allowed to the control-port.
#
# If you want to enable control-port access, comment-in the two following
# lines by removing the '#' sign at the beginning of each relevant line.
# Bare in mind that changes will be overwritten by subsequent updates.
#

set admin_user "nsadmin"
set admin_pass "x"

#
# Generic server configuration procedure. Generally, you should not
# change anything below, unless instructed by Archiware tech-support.
# Also, bare in mind that all potential changes will be overwritten
# by the subsequent updates.
#

proc srvconf {args} {

    global server name admin_user admin_pass

    set server vmsn

    #
    # Calculate the distribution home directory.
    # To do this, change to directory where the
    # configuration file is stored and get the
    # current working directory. This must be 
    # the "etc" under the installation directory.
    #

    set cf [ns_info config]

    set cwd [pwd]; cd [file dirname $cf]
    set cfd [pwd]; cd $cwd

    set home [file dirname $cfd]

    #
    # Test directory structure and create it if needed.
    #

    foreach dir {
        pages modules modules/tcl modules/nslog modules/nsperm modules/nsssl
    } {
        file mkdir $home/servers/$server/$dir
    }

    #
    # Global server parameters. There are myriad of server-related
    # parameters which one can put here. We use defaults for most,
    # and explicitly set only very few of them.
    #

    ns_section ns/parameters
    ns_param schedmaxelapsed 15
    ns_param home       $home
    ns_param tmpdir     $home/temp
    ns_param pidfile    log/$server.pid
    ns_param logroll    true
    ns_param serverlog  log/$server.log
    ns_param tcllibrary modules/tcl

    #
    # Rotate server log file per hand. It is really
    # tempting to use ns_logroll here but it will
    # not work always.
    #

    ns_rollfile log/$server.log 365

    #
    # Empirically determined minimum thread stacksize
    #

    ns_section ns/threads
    ns_param stacksize [expr 384 * 1024]

    #
    # MIME types. You can extend this section by adding additional
    # "ns_param" lines giving link between the file extension and
    # the corresponding mime type, like: ns_param ".gif" "image/gif".
    # Server has a reasonable number of common types already built-in.
    #

    ns_section ns/mimetypes

    ns_param noextension "*/*"
    ns_param default     "*/*"

    #
    # Supported servers.
    #

    ns_section ns/servers
    ns_param $server $server

    #
    # Server-specific parameters. There are myriad of server-related
    # parameters which one can put here. We use defaults for most.
    #

    ns_section ns/server/$server
    ns_param directoryfile  index.adp,index.html,index.htm
    ns_param globalstats    false
    ns_param urlstats       false
    ns_param maxconnections 128
    ns_param maxthreads     128
    ns_param threadtimeout  180
    ns_param maxpost        [expr {1024 * 1024 * 512}]

    ns_section ns/server/$server/tcl
    ns_param library  servers/$server/modules/tcl
    ns_param initfile lib/init.tcl

    #
    # Limit GUI handling threads per module.
    # Unfortunately, modules load too late
    # and there is no config-options there we
    # can modify, so we must do it here.
    #

    ns_section ns/server/$server/pools
    ns_param gui "GUI-service"

    ns_section ns/server/$server/pool/gui
    ns_param map "POST /login"
    ns_param map "GET  /login"
    ns_param map "POST /vmsnapp"
    ns_param map "GET  /vmsnapp"
    ns_param maxthreads 4
    ns_param minthreads 1
    ns_param threadtimeout 360

    #
    # HTTPS socket driver module parameters.
    #

    ns_section ns/server/$server/module/nsssl
    ns_param certificate servers/$server/modules/nsssl/server.pem
    ns_param maxinput [expr {1024 * 1024 * 512}]
    ns_param spoolerthreads 1
    ns_param address 0.0.0.0
    ns_param port 443

    #
    # Adjust default cache parameters for fastpath module.
    # Notice that same parameters are used for nsxcmd_cache.
    #

    ns_section ns/server/$server/fastpath
    ns_param pagedir       servers/$server/pages
    ns_param mmap          true
    ns_param cache         true
    ns_param cachemaxentry [expr {4096 * 1024}]

    #
    # Declare C-level/script NaviServer modules.
    # C-level ones are enumerated, script not.
    #

    ns_section ns/server/$server/modules

    foreach clib {
        xcmds
        yajltcl
        thread
        sv
        tea
        tls
        user
        volume
        sqlite
        zzip
        vmsncp
        tdomtdp.so
        nscp.so
        nslog.so
        nsproxy.so
        nsssl.so
        nsxcmd.so
    } {
        if {[file extension $clib] ne {}} {
            ns_param [file rootname $clib] $clib
        } else {
            ns_param $clib lib${clib}[info sharedlibextension]
        }
    }

    foreach dir [list $home/modules/tcl $home/servers/$server/modules/tcl] {
        foreach file [lsort [glob -nocomplain -types d -directory $dir *]] {
            ns_param [file tail $file] tcl
        }
    }

    #
    # Declare Tcl C-level modules which have
    # not (yet) been converted to NaviServer.
    #

    ns_section ns/server/$server/tclclibs

    foreach clib {
        nsf
    } {
        if {[file extension $clib] ne {}} {
            ns_param [file rootname $clib] $clib
        } else {
            ns_param $clib lib${clib}[info sharedlibextension]
        }
    }

    #
    # Control port parameters
    #

    if {[info exists admin_user] && [info exists admin_pass]} {
        ns_section ns/server/$server/module/nscp
        ns_param address 127.0.0.1
        ns_param port 1443
        ns_section ns/server/$server/module/nscp/users
        ns_param user $admin_user:[ns_crypt $admin_pass ab]:
    }

    #
    # Access log parameters.
    #

    ns_section ns/server/$server/module/nslog
    ns_param maxbackup    7
    ns_param rollhour     0
    ns_param rolllog      true
    ns_param rollonsignal true

    #
    # ADP (AOLserver Dynamic Page) parameters.
    # Note that we don't use ADP's but if we
    # don't declare it, we get the ugly message!
    #

    ns_section ns/server/$server/adp
    ns_param map /*.adp
}

#
# Now, run the configurator procedure
#

srvconf
