#!/usr/bin/env bash

# Archiware PURE console script.
#
# See the file "license.txt" for information on usage and
# redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.

###################### Test for root user ####################################

GROUP=`id -u`
if test $GROUP -ne 0; then
    echo "Archiware Pure can only be started by a root user!" >&2
    exit 1
fi
USER=`whoami`

###################### OS detection ##########################################

if [ -f "/etc/os-release" ]
then
    FLAVOR=$(grep "^ID=" "/etc/os-release" | cut -d'=' -f2)
else
    FLAVOR=$OSTYPE
fi

NASBOX=0
if test -e /bin/get_key_value && \
   test -e /etc/synoinfo.conf && \
   /bin/get_key_value /etc/synoinfo.conf unique | grep -q "synology"
then
    NASBOX=1
    FLAVOR=synology
fi

if test -e /etc/model.conf && \
   head /etc/model.conf | grep -q "VENDOR = QNAP"
then
    NASBOX=1
    FLAVOR=qnap
fi

if test -e /proc/readynas/model && \
   head  /proc/readynas/model | grep -q "ReadyNAS"
then
    NASBOX=1
    FLAVOR=readynas
fi

if_os () { [[ $OSTYPE == *$1* ]]; }
if_flavor () { [[ $FLAVOR == *$1* ]]; }

which whiptail && WINDOW=whiptail
which dialog && WINDOW=dialog
[ -z WINDOW ] && echo "Neither 'whiptail' nor 'dialog' packages are installed. \n
While these packages are not critical for Archiware Pure, some menu options in this script
won't be shown correctly. \n
Please install one of those packages."

window () {
    if [ -n $WINDOW ]
    then
        $WINDOW "$@"
    else
        echo "Warning. Neither whiptail nor dialog packages installed."
    fi
}

###################### init global vars ######################################

if_os bsd && export TERM=xterm

VMSNHOME="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
PARENT_DIR=$(dirname "$VMSNHOME")
PASS_FILE="$PARENT_DIR/password.set"
TEMP_DIR="$VMSNHOME/temp"
DATEFMT="%d-%h-%Y"
TIMEFMT="%T"

######################## Show_time function #################################

# Shows today's date and time

show_time() {
   window --title " Date & Time " \
   --msgbox "\n        Current Date:   `date +"$DATEFMT"` \n\n \
       Current Time:   `date +"$TIMEFMT"`" 10 50
}

############################ Date and Time Menu #############################

datetime() {
    while :
        do
            window --clear --title " Date and Time Properties " \
                   --menu "Use [UP/DOWN] key to move" 12 60 6 \
                "0" "Exit" \
                "1" "Display Date and Time" \
                "2" "Change Date" \
                "3" "Change Time" \
                "4" "Change Time Zone" 2> menuchoices.$$
            retopt=$?
            choice=`cat menuchoices.$$`
            case $retopt in
               0) case $choice in
                    1)  show_time ;;
                    2)  changedate ;;
                    3)  changetime ;;
                    4)  if_os linux && dpkg-reconfigure tzdata
                        if_os bsd && bsdconfig timezone
                        ;;
                    0)  return 0 ;;
                  esac ;;
              *) return 0 ;;
            esac
        done
}

enablentp() {
    #Enable auto date and time updating
    if_os linux && {
        EXIT_MSG=$( timedatectl set-ntp 1 )
        EXIT_STAT=${?}
        if [ "$EXIT_STAT" -ne "0" ] ; then
           window --title " Error Message  " --msgbox "\n $EXIT_MSG" 8 60
        else
           window --title " Info  " --msgbox "\n Automatically time and date updating successfully set" 8 60
        fi
    }
}

changedate() {

    if_os linux && {
        TXT_YES="Chose \ZbYES\ZB if you would like to use automatically date and time updating"
        TXT_NO="Chose \ZbNO\ZB if you would like to set date and time manually"

        window --title " Automaticaly date updating  " --yesno "\n $TXT_YES \n\n $TXT_NO" 10 40
        case $? in
            #Yes
            0)
                enablentp
                ;;
            #No
            1)
                #Disable auto date and time updating
                EXIT_MSG=$( timedatectl set-ntp 0 )
                TODAY=$( date +"%d %m %Y" )
                DATUM=$(window --title " Date configuration " --inputbox "Please enter a date in dd-mm-YYYY format:" 0 0 $TODAY 3>&1 1>&2 2>&3 3>&-)
                if [ "$?" -ne "0" ]; then
                    return 0
                fi
                window --title " Confirmation " --yesno "\n\n       Set date: $DATUM" 10 40
                if [ "$?" -eq "0" ] ; then
                    DATREFM=$(date -d $DATUM +"%d-%m-%Y")
                    CURTIME=$(date +"$TIMEFMT")
                    EXIT_MSG=$( timedatectl set-time "$DATREFM $CURTIME" 3>&1 1>&2 2>&3 3>&-)
                    EXIT_STAT=${?}
                    if [ "$EXIT_STAT" -ne "0" ] && [ "$EXIT_STAT" -ne "2" ] ; then
                        window --title " Error Message " --msgbox "\n  $EXIT_MSG" 8 60
                    else
                        show_time
                    fi
                fi
                ;;
        esac
    }

    if_os bsd && {
        TODAY=date +"$datefmt"
        DATUM=$(window --title " Date configuration " --inputbox "Please enter a date in dd-mm-YYYY format:" 0 0 $TODAY 3>&1 1>&2 2>&3 3>&-)
        if [ "$?" -ne "0" ]; then
            return 0
        fi
        window --title " Confirmation " --yesno "\n\n       Set date: $DATUM" 10 40
        if [ "$?" -eq "0" ] ; then
            EXIT_MSG=$( date -f "$DATEFMT" $DATUM 3>&1 1>&2 2>&3 3>&-)
            EXIT_STAT=${?}
            if [ "$EXIT_STAT" -ne "0" ] && [ "$EXIT_STAT" -ne "2" ] ; then
                window --title " Error Message " --msgbox "\n  $EXIT_MSG" 8 60
            else
                show_time
            fi
        fi   
    }
}

changetime() {
    if_os linux && {
        TXT_YES="Do you want to enable automatic date and time updates?"
    
        window --title " Automatic date time updates  " --yesno "$TXT_YES" 10 40
        case $? in
            #Yes
            0)
                enablentp
            ;;
        
            #No
            1)
                #Disable auto date and time updating
                EXIT_MSG=$( timedatectl set-ntp 0 )
        
                TODAY=$( date +"%H %M %S" )
                DATUM=$(window --title " Time " --inputbox "Enter time in 24h format 'hh mm ss':" 4 40 $TODAY 3>&1 1>&2 2>&3 3>&-)
                if [ "$?" -ne "0" ]; then
                    return 0
                fi
                window --title " Confirmation " --yesno "\n\n       Set time: $DATUM" 10 40
                if [ "$?" -eq "0" ] ; then
                   CURDATE=$(date +"%Y-%m-%d")
                   EXIT_MSG=$( timedatectl set-time "$CURDATE $DATUM" 3>&1 1>&2 2>&3 3>&-)
                   EXIT_STAT=${?}
                   if [ "$EXIT_STAT" -ne "0" ] && [ "$EXIT_STAT" -ne "2" ] ; then
                      window --title " Error Message " --msgbox "\n  $EXIT_MSG" 8 60
                   else
                      show_time
                   fi
                fi
            ;;
        esac
    }

    if_os bsd && {
        TODAY=date +"$TIMEFMT"
        DATUM=$(window --title " Time " --inputbox "Enter time in 24h format 'hh mm ss':" 4 40 $TODAY 3>&1 1>&2 2>&3 3>&-)
        if [ "$?" -ne "0" ]; then
            return 0
        fi
        window --title " Confirmation " --yesno "\n\n       Set date: $DATUM" 10 40
        if [ "$?" -eq "0" ] ; then
            EXIT_MSG=$( date -f "$TIMEFMT" $DATUM 3>&1 1>&2 2>&3 3>&-)
            EXIT_STAT=${?}
            if [ "$EXIT_STAT" -ne "0" ] && [ "$EXIT_STAT" -ne "2" ] ; then
                window --title " Error Message " --msgbox "\n  $EXIT_MSG" 8 60
            else
                show_time
            fi
        fi   
    }
}

############################### Start SSH Server #############################

start_sshd(){
	echo "Starting SSH Server..."
	if_os linux && EXIT_MSG=$( systemctl start ssh )
    if_os bsd && EXIT_MSG=$( /etc/rc.d/sshd onestart )
	EXIT_STAT=${?}

	if [[ "$EXIT_STAT" -eq "0" ]]; then
	    echo "$EXIT_MSG"
	    echo "SSH Server is UP and running!"
	fi
}

############################### Start SSH Server #############################

stop_sshd(){
	echo "Stopping SSH Server..."
	if_os linux && EXIT_MSG=$( systemctl stop ssh )
    if_os bsd && EXIT_MSG=$( /etc/rc.d/sshd stop )
	EXIT_STAT=${?}

	if [[ "$EXIT_STAT" -eq "0" ]]; then
	    echo "$EXIT_MSG"
	    echo "SSH Server is DOWN!"
	fi
}

############################### restart function #############################

do_restart() {

    window --title " Confirmation " --defaultno --yesno "\n    Do you really want to restart Archiware Pure software? (y/n)" 8 50
    if [ "$?" -eq "0" ] ; then
        $VMSNHOME/etc/restart.sh
    else
        return 1
    fi

}

############################### reboot function #############################

do_reboot() {

    window --title " Confirmation " --defaultno --yesno "\n    Do you really want to reboot? (y/n)" 8 50
    if [ "$?" -eq "0" ] ; then
        EXIT_MSG=$( reboot 3>&1 1>&2 2>&3 3>&- )
        EXIT_STAT=${?}
        if [ "$EXIT_STAT" -ne "0" ] ; then
            window --title " Error Message " --msgbox "\n  $EXIT_MSG" 8 70
        else
            exit
        fi
    else
        return 1
    fi

}

############################### shutdown function #############################

do_shutdown() {

    window --title " Confirmation " --defaultno --yesno "\n    Do you really want to shutdown? (y/n)" 8 50
    if [ "$?" -eq "0" ] ; then
        EXIT_MSG=$( poweroff 3>&1 1>&2 2>&3 3>&- )
        EXIT_STAT=${?}
        if [ "$EXIT_STAT" -ne "0" ] ; then
            window --title " Error Message " --msgbox "\n  $EXIT_MSG" 8 70
        else
            exit
        fi
    else
        return 1
    fi
}

############################### password reset #############################

passwd_reset() {
    clear
    PASSWD1=$(window --passwordbox "Please enter the new password:" 10 40 3>&1 1>&2 2>&3 3>&-)
    if [ "$?" -ne "0" ]; then return 0; fi
    PASSWD2=$(window --passwordbox "Please repeat the new password:" 10 40 3>&1 1>&2 2>&3 3>&-)
    if [ "$?" -ne "0" ]; then return 0; fi

    if [ $PASSWD1 == $PASSWD2 ]; then
        EXIT_MSG1=$( echo -e "$PASSWD1\n$PASSWD2" | passwd $USER 3>&1 1>&2 2>&3 3>&- )
        EXIT_STAT1=${?}
        EXIT_MSG2=$( echo -e "$PASSWD1\n$PASSWD2" | passwd pure 3>&1 1>&2 2>&3 3>&- )
        EXIT_STAT2=${?}
        if [ "$EXIT_STAT1" -ne "0" ] ; then
            window --title " Error Message " --msgbox "\n  $EXIT_MSG1" 8 70
        elif [ "$EXIT_STAT2" -ne "0" ] ; then
            window --title " Error Message " --msgbox "\n  $EXIT_MSG2" 8 70
        else
            rm $PASS_FILE
            window --msgbox "\n  Password successfully changed." 8 40
        fi
    else
        window --title " Warning " --msgbox "\n   Passwords do not match. Please try again." 8 50
    fi
}

############################### configure network ############################

network_configure() {

    clear
    MANAGER=$(which nmtui)
    if [ -z $MANAGER ]
    then
        window --title " Error Message " --msgbox "Assisted network configuration requires 'network-manager' package.
        Please install the package or configure netwrok settings manually." 8 70
    else
        ($MANAGER)
    fi
}

############################### get server url #############################

print_urls() {

    if_os linux && IP=$(ip -4 addr show scope global | awk '$1 == "inet" {print $2}' | cut -d'/' -f1)
    if_os linux && PORT=$(ss -lnp | grep nsd | grep -v 127.0.0.1 | awk {'print $5'} | cut -d':' -f2)
    echo "https://$IP:$PORT"
}

############################## main menu ####################################

while :
do
    clear
	if [ -f $PASS_FILE ]
    then
        passwd_reset
        continue
    fi
    
    clear

    PUREWELCOME="Welcome to ARCHIWARE PURE Server Appliance"
    PUREVER=$(cat "$VMSNHOME"/log/vmsn.log 2> /dev/null | grep "Software Version:" | awk '{print $5}')

    echo
    echo -e "\e[1m $PUREWELCOME\e[0m"
    echo
    echo " PURE Version:" "$PUREVER"
    echo " OS Appliance:" `uname -o` `uname -r`
    echo
    echo " To manage your appliance please browse to:"
    for line in $(print_urls) ; do
       echo -e "    \e[1;3;92m$line\e[0m"
    done

    # display menu
    echo
    echo
    echo " PURE Console Menu"
    echo " -----------------"
    echo " 1) Reset Password"
    echo " 2) Configure Network"
    echo " 3) Date and Time Properties"
    echo " 4) Start SSH Server"
    echo " 5) Stop SSH Server"
    echo
    echo " p) Restart Archiware Pure"
    echo " r) Reboot Pure VM"
    echo " s) Shut down Pure VM"
    echo " x) Exit to shell"

    read -p " Enter a number: " OPMODE        
    case $OPMODE in
        1) passwd_reset ;;
        2) network_configure ;;
        3) datetime ;;
        4) start_sshd ;;
        5) stop_sshd ;;
        p) do_restart ;;
        r) do_reboot ;;
        s) do_shutdown ;;
        x) echo "Exiting... To run the console menu again, execute"
           echo "$VMSNHOME/vmsn_console.sh" 
           exit ;;
        *) clear ;;
    esac

done
